/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

//import Dropdown from "./dropdown.js"; // dropdown is currently implemented using Bootstrap
import hyperHTML from "./node_modules/hyperhtml/esm/index.js";

/** @typedef {object} PrivateData
 * @property {ShadowRoot} shadow
 * @property {HTMLImageElement} icon
 * @property {HTMLLabelElement} label
 * @property {HTMLElement} accel
 * @property {boolean} command_set
 */

/** @type {WeakMap<MenuItemElement,PrivateData>} */
const PRIVATE_DATA = new WeakMap();
const TAG_NAME = "menu-item";

const PCENChar =
	'[-._a-z0-9\\xB7\\xC0-\\xD6\\xD8-\\xF6\\xF8-\\u037D\\u037F-\\u1FFF\\u200C-\\u200D\\u203F-\\u2040' +
	'\\u2070-\\u218F\\u2C00-\\u2FEF\\u3001-\\uD7FF\\uF900-\\uFDCF\\uFDF0-\\uFFFD\\u{10000}-\\u{EFFFF}]';

let registered = false;

class MenuItemElement extends HTMLElement {
	static get VALID_TYPES() {
		return [
			"checkbox",
			"command",
			"radio"
		];
	}

	static set TAG_NAME(newTagName) {
		if (registered) {
			throw "Can’t change the tag name once registered";
		}
		if (!new RegExp(`^[a-z]${PCENChar}*-${PCENChar}$`,'u').test(newTagName)) {
			throw `The specified tag name "${newTagName}" contains invalid characters `;
		}
		TAG_NAME = newTagName;
	}

	static get TAG_NAME() {
		return TAG_NAME;
	}

	static get observedAttributes() {
		return [
			"_moz-accel",
			"command",
			"icon",
			"label",
			"type"
		];
	}

	constructor() {
		super();
		/** @type {PrivateData} */
		const _private = {};
		PRIVATE_DATA.set(this, _private);
		_private.command_set = false;

		const shadow = _private.shadow = this.attachShadow({mode:'closed'});

		hyperHTML(shadow)`
			<link rel="stylesheet" type="text/css" href="./menu-item.css"/>
			<div class="icon">
				<img id="icon"/>
			</div>
			<label id="label"/>
			<div id="accel"/>
			<div class="menu-right">
				<slot name="menu-slot"/>
			</div>`;

		_private.icon = shadow.getElementById("icon");
		_private.label = shadow.getElementById("label");
		_private.accel = shadow.getElementById("accel");

		this.tabIndex = 0;
	}

	connectedCallback() {
		this.addEventListener("click", (evt) => {
			switch (this.type) {
				case "command":
					if (this.command) {
						//this.command.click();
						this.command.dispatchEvent(evt);
					}
					break;
				case "checkbox":
					this.checked = !this.checked;
					break;
				case "radio":
					if (this.radiogroup) {
						for (const btn of document.querySelectorAll(`menu-item[type="radio"]`)) {
							if (btn !== this && btn.radiogroup === this.radiogroup) {
								btn.checked = false;
								this.checked = true;
							}
						}
					}
					this.checked = true;
					break;
				case "menu":

			}
		});

		this.addEventListener("mouseenter", (evt) => {
			for (const c of this.parentElement.children) {
				if (c !== this && c instanceof MenuItemElement) {
					c.removeAttribute("_moz-menuactive");
				}
			}
			this.setAttribute("_moz-menuactive",true);
		});

		this.addEventListener("mouseleave", (evt) => {
			if (this.type === "menu") {
				for (const c of this.menu.querySelectorAll(`${TAG_NAME}[_moz-menuactive]`)) {
					c.removeAttribute("_moz-menuactive");
				}
			}
		});

		this.setAttribute("aria-role","menuitem");
	}

	disconnectedCallback() {}

	attributeChangedCallback(name, oldValue, newValue) {
		switch (name) {
			case "_moz-accel":
				PRIVATE_DATA.get(this).accel.textContent = newValue;
				break;
			case "command":
				if (!PRIVATE_DATA.get(this).command_set) {
					const command = document.getElementById(newValue);
					if (command) {
						PRIVATE_DATA.get(this).command = command;
					}
				}
				break;
			case "icon":
				PRIVATE_DATA.get(this).icon.src = newValue;
				break;
			case "label":
				PRIVATE_DATA.get(this).label.textContent = newValue;
				break;
			case "type":
				switch (newValue) {
					case "menu":
						let newLabel = this.menu.hasAttribute("label") ? this.menu.getAttribute("label") : null;
						if (newLabel) {
							PRIVATE_DATA.get(this).label.textContent = newLabel;
						} else if (this.hasAttribute("label")) {
							PRIVATE_DATA.get(this).label.textContent = this.getAttribute("label");
						}
						this.menu.setAttribute("slot", "menu-slot");
						break;
				}
				break;
		}
	}

	/**
	 * @return {HTMLMenuElement|undefined} The menu element,
	 *	or `undefined` if this menu item is not of type menu.
	 */
	get menu() {
		if (this.getAttribute("type") === "menu") {
			return this.querySelector("menu");
		}
		return undefined;
	}

	/** @return	{string}	*/ get icon()	{ return this.getAttribute("icon");	}
	/** @return	{string}	*/ get radiogroup()	{ return this.getAttribute("radiogroup");	}
	/** @return	{string}	*/
	get type()	{
		if (!this.hasAttribute("type")) return "command";
		let type = this.getAttribute("type").toLowerCase();
		if (MenuItemElement.VALID_TYPES.includes(type)) return type;
		if (type === "menu" && this.menu) {
			return type;
		}
		return "command";
	}

	/** @return	{boolean}	*/ get disabled()	{ return this.hasAttribute("disabled");	}
	/** @return	{boolean}	*/ get checked()	{ return this.hasAttribute("checked");	}
	/** @return	{boolean}	*/ get default()	{ return this.hasAttribute("default");	}

	/** @param	{string}	type	*/ set type(type)	{ this.setAttribute("type", type);	}
	/** @param	{string}	icon	*/ set icon(icon)	{ this.setAttribute("icon", icon);	}
	/** @param	{string}	radiogroup	*/ set radiogroup(radiogroup)	{ this.setAttribute("radiogroup", radiogroup);	}

	/** @param	{boolean}	disabled	*/
	set disabled(disabled) {
		if (disabled) this.setAttribute("disabled", "");
		else this.removeAttribute("disabled");
	}
	/** @param	{boolean}	checked	*/
	set checked(checked) {
		if (checked) this.setAttribute("checked", "");
		else this.removeAttribute("checked");
	}
	/** @param	{boolean}	_default	*/
	set default(_default) {
		if (_default) this.setAttribute("default", "");
		else this.removeAttribute("default");
	}

	/** @returns {Element|undefined} */
	get command() {
		const _command = PRIVATE_DATA.get(this).command;
		if (_command) return _command;
		if (this.hasAttribute("command")) {
			return document.getElementById(this.getAttribute("command"));
		}
		return undefined;
	}

	/**
	 * @param {EventTarget|undefined} command
	 */
	set command(command) {
		if (command instanceof EventTarget) {
			PRIVATE_DATA.get(this).command = command;
			PRIVATE_DATA.get(this).command_set = true;
			if (command instanceof Element) {
				this.setAttribute("command", command.id);
			} else {
				this.removeAttribute("command");
			}
		} else {
			PRIVATE_DATA.get(this).command = undefined;
			PRIVATE_DATA.get(this).command_set = true;
			this.setAttribute("command", command);
		}
		PRIVATE_DATA.get(this).command_set = false;
	}

	/**
	 * @return {string} The label
	 */
	get label() {
		if (this.getAttribute("type") === "menu" && this.menu.hasAttribute("label")) {
			return this.menu.getAttribute("label");
		}
		return this.getAttribute("label");
	}

	/**
	 * @param {string} label The label
	 */
	set label(label) {
		if (this.getAttribute("type") === "menu") {
			this.menu.setAttribute("label", label);
			PRIVATE_DATA.get(this).label.textContent = label;
		} else {
			this.setAttribute("label", label);
		}
	}
}

const loadPromise = new Promise((resolve) => {
	document.addEventListener('DOMContentLoaded', () => {
		customElements.define(TAG_NAME, MenuItemElement);
		registered = true;
		resolve(MenuItemElement);
	});
});

export default MenuItemElement;
